/*
 Copyright (c) Information Equipment co.,LTD.
 All right reserved.
 
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com> Made in KOREA

 http://minzkn.pe.ky
*/

#ifndef DEF_SOURCE_buffer_c 
#define DEF_SOURCE_buffer_c "buffer.c"

#ifdef WIN32
#include <stdio.h>
#include <windows.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

#include "mz.h"

#define DEF_BUFFER_ALIGN         (8 << 10) /* Default allocate block size */
#define DEF_BUFFER_MAX_ALIGN     (DEF_BUFFER_ALIGN << 4) /* Max allocate block size */

#ifndef DEF_MACRO_PEEK
#define DEF_MACRO_PEEK (1)
#define __MZ_PEEK__(s_From, s_Offset) (((unsigned char *)(s_From)) + (s_Offset))
#define __MZ_PEEK_R__(s_From, s_Offset) (((unsigned char *)(s_From)) - (s_Offset))
#endif

#ifndef DEF_FUNCTION_ALLOC
#define DEF_FUNCTION_ALLOC (1)
#define MZ_Alloc malloc
#define MZ_Free free
#endif

#ifndef DEF_FUNCTION_MEMORY
#define DEF_FUNCTION_MEMORY (1)
#define MZ_ZeroMemory(s_To, s_Size) memset((s_To), 0, (s_Size))
#define MZ_SetMemory(s_To, s_Value, s_Size) memset((s_To), (s_Value), (s_Size))
#define MZ_CopyMemory(s_To, s_From, s_Size) memcpy((s_To), (s_From), (s_Size))
#endif

#ifndef DEF_STRUCT_BUFFER
#define DEF_STRUCT_BUFFER (1)
typedef struct ts_MZ_Buffer_Node
{
 struct ts_MZ_Buffer_Node *Next;
 void *Data;
 int Index, PopIndex, Size;
}t_MZ_Buffer_Node;

typedef struct ts_MZ_Buffer
{
 t_MZ_Buffer_Node *Node, *NewNode, *LastNode;
 int Size;
}t_MZ_Buffer;
#endif

t_MZ_Buffer *MZ_CreateBuffer(void);
t_MZ_Buffer_Node *MZ_DestroyBuffer_Node(t_MZ_Buffer_Node *s_Buffer_Node);
t_MZ_Buffer *MZ_DestroyBuffer(t_MZ_Buffer *s_Buffer);
int MZ_CheckBuffer(t_MZ_Buffer *s_Buffer);
int MZ_LimitCheckBuffer(t_MZ_Buffer *s_Buffer, int s_LimitSize);
t_MZ_Buffer *MZ_PushBuffer(t_MZ_Buffer *s_Buffer, const void *s_Data, int s_Size);
t_MZ_Buffer *MZ_PassPushBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size);
int MZ_PopBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size);
int MZ_PrePopBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size, int s_Offset);
void *MZ_PassPopBuffer(t_MZ_Buffer *s_Buffer, int *s_Size);
t_MZ_Buffer *MZ_FlushBuffer(t_MZ_Buffer *s_Buffer);

/* [ Static function ] ==================================================================== */

/* [ Public function ] ==================================================================== */
t_MZ_Buffer *MZ_CreateBuffer(void)
{
 t_MZ_Buffer *s_Return;
 s_Return = (t_MZ_Buffer *)MZ_Alloc(sizeof(t_MZ_Buffer));
 if(s_Return)
 {
  s_Return->Node = s_Return->NewNode = s_Return->LastNode = (t_MZ_Buffer_Node *)0;
  s_Return->Size = 0;
 }
 return(s_Return);
}

t_MZ_Buffer_Node *MZ_DestroyBuffer_Node(t_MZ_Buffer_Node *s_Buffer_Node)
{
 t_MZ_Buffer_Node *s_Node;
 while(s_Buffer_Node)
 {
  s_Node = s_Buffer_Node;
  s_Buffer_Node = s_Buffer_Node->Next;
  if(s_Node->Data)MZ_Free((void *)(s_Node->Data));
  MZ_Free((void *)(s_Node));
 }
 return(s_Buffer_Node);
}

t_MZ_Buffer *MZ_DestroyBuffer(t_MZ_Buffer *s_Buffer)
{
 if(s_Buffer)
 {
  s_Buffer->Node = MZ_DestroyBuffer_Node(s_Buffer->Node);
  s_Buffer->NewNode = MZ_DestroyBuffer_Node(s_Buffer->NewNode);
  MZ_Free((void *)(s_Buffer));
  s_Buffer = (t_MZ_Buffer *)0;
 }
 return(s_Buffer);
}

int MZ_CheckBuffer(t_MZ_Buffer *s_Buffer)
{
 if(s_Buffer)return(s_Buffer->Size);
 return(0);
}

int MZ_LimitCheckBuffer(t_MZ_Buffer *s_Buffer, int s_LimitSize)
{
 if(s_Buffer)
 {
  if(s_Buffer->Size < s_LimitSize)return(1);
 }
 return(0);
}

t_MZ_Buffer *MZ_SyncBuffer(t_MZ_Buffer *s_Buffer)
{
 if(s_Buffer)
 {
  if(s_Buffer->NewNode)
  {  
   if(s_Buffer->Node)
   {
    s_Buffer->LastNode->Size = s_Buffer->LastNode->Index; /* Strip last buffer */
    s_Buffer->LastNode->Next = s_Buffer->NewNode;
    s_Buffer->LastNode = s_Buffer->LastNode->Next;	
   }
   else s_Buffer->LastNode = s_Buffer->Node = s_Buffer->NewNode;
   s_Buffer->NewNode = (t_MZ_Buffer_Node *)0; /* Passed */
  }
 }
 return(s_Buffer);
}

t_MZ_Buffer *MZ_PushBuffer(t_MZ_Buffer *s_Buffer, const void *s_Data, int s_Size)
{
 t_MZ_Buffer_Node *s_Node;
 int s_PushedSize, s_PushSize;
 if(s_Buffer == (t_MZ_Buffer *)0)s_Buffer = MZ_CreateBuffer();
 if(s_Buffer && s_Size > 0)
 {
  s_PushedSize = 0;
  if(s_Buffer->LastNode)
  { /* Already node push */   
   s_Node = s_Buffer->LastNode;   
   if(s_Node->Data == (void *)0)
   {
    if((s_Size - s_PushedSize) < DEF_BUFFER_ALIGN || (s_Size - s_PushedSize) > DEF_BUFFER_MAX_ALIGN)s_Node->Size = DEF_BUFFER_ALIGN;    
    else s_Node->Size = (s_Size - s_PushedSize);
    s_Node->Data = (void *)MZ_Alloc(s_Node->Size);
    s_Node->Index = s_Node->PopIndex = 0;
   }
   if(s_Node->Index < s_Node->Size)
   {
    s_PushSize = s_Node->Size - s_Node->Index;
    if(s_PushSize > (s_Size - s_PushedSize))s_PushSize = (s_Size - s_PushedSize);
    if(s_PushSize > 0)
    {
     if(s_Data)MZ_CopyMemory((void *)__MZ_PEEK__(s_Node->Data, s_Node->Index), (void *)__MZ_PEEK__(s_Data, s_PushedSize), s_PushSize);
     else MZ_ZeroMemory((void *)__MZ_PEEK__(s_Node->Data, s_Node->Index), s_PushSize);	
     s_Node->Index += s_PushSize;
     s_Buffer->Size += s_PushSize;
     s_PushedSize += s_PushSize;
    }
   }
  }
  while(s_PushedSize < s_Size)
  { /* New node push */
   if(s_Buffer->NewNode == (t_MZ_Buffer_Node *)0)
   {
    s_Node = (t_MZ_Buffer_Node *)MZ_Alloc(sizeof(t_MZ_Buffer_Node));
    if(s_Node)
    {
     s_Node->Next = (t_MZ_Buffer_Node *)0;	 
     s_Node->Index = s_Node->PopIndex = 0;
     if((s_Size - s_PushedSize) < DEF_BUFFER_ALIGN || (s_Size - s_PushedSize) > DEF_BUFFER_MAX_ALIGN)s_Node->Size = DEF_BUFFER_ALIGN;    
     else s_Node->Size = (s_Size - s_PushedSize);	 
     s_Node->Data = (void *)MZ_Alloc(s_Node->Size);
     if(s_Node->Data)s_Buffer->NewNode = s_Node;
     else
     { /* Wait for free memory */	  
      (void)MZ_DestroyBuffer_Node(s_Node);
      sleep(1); /* MZ_mSleep(100); */
     }
    }
    else sleep(1); /* MZ_mSleep(100); */  /* Wait for free memory */
   }
   if(s_Buffer->NewNode)
   {
    s_Node = s_Buffer->NewNode;
    s_PushSize = s_Node->Size - s_Node->Index;
    if(s_PushSize > (s_Size - s_PushedSize))s_PushSize = (s_Size - s_PushedSize);
    if(s_PushSize > 0)
    {
     if(s_Data)MZ_CopyMemory((void *)__MZ_PEEK__(s_Node->Data, s_Node->Index), (void *)__MZ_PEEK__(s_Data, s_PushedSize), s_PushSize);
     else MZ_ZeroMemory((void *)__MZ_PEEK__(s_Node->Data, s_Node->Index), s_PushSize);
     s_Node->Index += s_PushSize;
     s_Buffer->Size += s_PushSize;
     s_PushedSize += s_PushSize;
    }	
    if(s_Node->Index == s_Node->Size)s_Buffer = MZ_SyncBuffer(s_Buffer);
   }
  }
 }
 return(s_Buffer);
}

t_MZ_Buffer *MZ_PassPushBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size)
{
 t_MZ_Buffer_Node *s_Node;
 if(s_Buffer == (t_MZ_Buffer *)0)s_Buffer = MZ_CreateBuffer();
 if(s_Buffer && s_Size > 0)
 {
  if(s_Buffer->NewNode)s_Buffer = MZ_SyncBuffer(s_Buffer);
  do
  {
   s_Node = (t_MZ_Buffer_Node *)MZ_Alloc(sizeof(t_MZ_Buffer_Node));
   if(s_Node)
   {
    s_Node->Next = (t_MZ_Buffer_Node *)0;
    s_Node->Index = s_Node->Size = s_Size;
    s_Node->PopIndex = 0;
    if(s_Data)s_Node->Data = (void *)s_Data;
    else s_Node->Data = (void *)MZ_Alloc(s_Size);
    if(s_Node->Data)
    {
     s_Buffer->NewNode = s_Node;
     s_Buffer = MZ_SyncBuffer(s_Buffer);
    }
    else 
    { /* Wait for free memory */
     s_Node = MZ_DestroyBuffer_Node(s_Node);
     sleep(1); /* MZ_mSleep(100);  */
    }	
   }
   else sleep(1); /* MZ_mSleep(100); */ /* Wait for free memory */
  }while(s_Node); /* Wait for free memory loop */
 }
 return(s_Buffer);
}

int MZ_PopBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size)
{
 int s_Return = 0, s_PopSize; /* Poped size */ 
 t_MZ_Buffer_Node *s_Node;
 if(s_Buffer)
 { 
  if(s_Buffer->NewNode)
  { /* More buffer sync */
   if((s_Buffer->Size - (s_Buffer->NewNode->Index - s_Buffer->NewNode->PopIndex)) < (s_Size - s_Return) && s_Buffer->NewNode->Data)s_Buffer = MZ_SyncBuffer(s_Buffer);
  }
  s_Node = s_Buffer->Node;
  while(s_Return < s_Size && s_Node)
  {    
   s_PopSize = s_Size - s_Return;
   if(s_PopSize > (s_Node->Index - s_Node->PopIndex))s_PopSize = (s_Node->Index - s_Node->PopIndex);   
   if(s_PopSize > 0)
   { /* Pop process */
    if(s_Data)MZ_CopyMemory((void *)(__MZ_PEEK__(s_Data, s_Return)), (void *)__MZ_PEEK__(s_Node->Data, s_Node->PopIndex), s_PopSize);
    s_Buffer->Size -= s_PopSize;
    s_Node->PopIndex += s_PopSize;
    s_Return += s_PopSize;
   }      
   if(s_Node->PopIndex >= s_Node->Index)
   {    
    s_Buffer->Node = s_Buffer->Node->Next;   
    s_Node->Next = (t_MZ_Buffer_Node *)0;
    (void)MZ_DestroyBuffer_Node(s_Node);
    s_Node = s_Buffer->Node;	
   }
  }
  if(s_Buffer->Node == (t_MZ_Buffer_Node *)0)s_Buffer->LastNode = (t_MZ_Buffer_Node *)0;
 }
 return(s_Return);
}

int MZ_PrePopBuffer(t_MZ_Buffer *s_Buffer, void *s_Data, int s_Size, int s_Offset)
{
 int s_Return = 0, s_PopSize, s_MyPopIndex; /* Poped size */ 
 t_MZ_Buffer_Node *s_Node;
 if(s_Buffer)
 { 
  if(s_Buffer->NewNode)
  { /* More buffer sync */
   if((s_Buffer->Size - (s_Buffer->NewNode->Index - s_Buffer->NewNode->PopIndex)) < (s_Size - s_Return) && s_Buffer->NewNode->Data)s_Buffer = MZ_SyncBuffer(s_Buffer);
  }
  s_Node = s_Buffer->Node;
  s_MyPopIndex = 0;
  while(s_Offset > 0 && s_Node)
  {
   s_MyPopIndex = s_Node->PopIndex;
   s_PopSize = (s_Node->Index - s_MyPopIndex);
   if(s_Offset <= s_PopSize)s_PopSize = s_Offset;
   s_Offset -= s_PopSize;
   s_MyPopIndex += s_PopSize;
   if(s_MyPopIndex >= s_Node->Index)
   {
    s_Node = s_Node->Next;
    if(s_Node)s_MyPopIndex = s_Node->PopIndex;
   }
  }  
  while(s_Return < s_Size && s_Node)
  {   
   s_PopSize = s_Size - s_Return;
   if(s_PopSize > (s_Node->Index - s_MyPopIndex))s_PopSize = (s_Node->Index - s_MyPopIndex);   
   s_PopSize = (s_Node->Index - s_MyPopIndex);
   if((s_Size - s_Return) <= s_PopSize)s_PopSize = (s_Size - s_Return);
   if(s_PopSize > 0)
   { /* Pop process */   
    if(s_Data)MZ_CopyMemory((void *)(__MZ_PEEK__(s_Data, s_Return)), (void *)__MZ_PEEK__(s_Node->Data, s_MyPopIndex), s_PopSize);	
    s_MyPopIndex += s_PopSize;		
    s_Return += s_PopSize;
   }      
   if(s_MyPopIndex >= s_Node->Index)
   {
    s_Node = s_Node->Next;   
    if(s_Node)s_MyPopIndex = s_Node->PopIndex;
   }   
  }  
 }
 return(s_Return);
}

void *MZ_PassPopBuffer(t_MZ_Buffer *s_Buffer, int *s_Size)
{
 void *s_Return = (void *)0; 
 t_MZ_Buffer_Node *s_Node;
 if(s_Size)*(s_Size) = 0; 
 if(s_Buffer)
 { 
  if(s_Buffer->Node)
  {
   s_Node = s_Buffer->Node;
   s_Buffer->Node = s_Buffer->Node->Next;      
   if(s_Node->Data && s_Node->Index > 0)
   {
    if(s_Size)*(s_Size) = s_Node->Index;
    s_Return = s_Node->Data;
    s_Node->Data = (void *)0;
   }
   s_Node->Next = (t_MZ_Buffer_Node *)0;
   (void)MZ_DestroyBuffer_Node(s_Node);
  }
 }  
 return(s_Return);
}

t_MZ_Buffer *MZ_FlushBuffer(t_MZ_Buffer *s_Buffer)
{
 if(s_Buffer)MZ_PopBuffer(s_Buffer, (void *)0, MZ_CheckBuffer(s_Buffer)); 
 return(s_Buffer);
}

#endif

/* End of source */
